------------------------------------------------------------------------
--    Redundancy Analyzer
--    by Mara Alpuente, Santiago Escobar and Salvador Lucas
--    Version of 17/10/2003 (Universidad Politcnica de Valencia)
------------------------------------------------------------------------

------------------------------------
-- Analysis of redundant arguments
------------------------------------

import erasureRedArgs
import List
import Flat
import interpreter

showVersion = do
  putStrLn "*************************************************************"
  putStrLn "**********Redundancy Analyzer and Erasure Procedure**********"
  putStrLn "*Universidad Politecnica de Valencia (Version of 16/11/2001)*"
  putStrLn "*************************************************************\n"

-------------------------------------------------
-- Complete Analysis
analysisRedArgs file = do
  showVersion
  putStrLn (">>>Reading and Compiling program "++file++"\n")
  prog <- readFlatCurry file
  if not (orthoProg prog) then putStrLn ">>>ERROR: The TRS is not orthogonal\n" else
   if not (testCDProg prog) then putStrLn ">>>ERROR: The TRS is not completely defined\n" else do
    putStrLn "\n>>>Starting redundant argument analysis..."
    putStrLn "!!!!!! Note: the TRS must be terminating !!!!!!"
    RArgs <- return (redArgs prog)
    putStrLn ("\n>>Redundant arguments detected:"++ppRedArgs RArgs)
    eraseFile RArgs file
 where 
  ppRedArgs [] = "\n"
  ppRedArgs ((EFunc F Args):xs) = "\n" ++ F ++ ": " ++ show Args ++ ppRedArgs xs

------------------------------------------------------------------------
-- Analysis of Redundant Arguments

------------------
-- First approach

-- Redundant Variable
--  1) No appearing in Exp
--  2) Within sub-constructor positions of 
--     the i-th parameter of f-rooted subterms of Exp
redVar _ _ V (Var VarIndex) = VarIndex /= V
redVar _ _ _ (Lit _) = False
redVar F I V (Apply E1 E2) = all (redVar F I V) [E1,E2]
redVar F I V (Constr _ Expr) = redVar F I V Expr
redVar F I V (Or E1 E2) = all (redVar F I V) [E1,E2]
redVar F I V (Case _ _ BE) = all (redVar F I V) (map (\(Branch _ x)->x) BE)
redVar F I V (Choice Expr) = redVar F I V Expr
redVar F I V (GuardedExpr _ E1 E2) = all (redVar F I V) [E1,E2]
redVar F I V (Comb Call FCall Exprs) = 
  ( is (\(FuncCall)->True) Call && FCall==F &&
    all (redVar F I V) (dropN I Exprs) && subConsPos V (getN I Exprs)  )
  || all (redVar F I V) Exprs
 
subConsPos V E 
  | is (\(Var _)->True) E             = subConsVar V E
  | is (\(Comb ConsCall _ _)->True) E = subConsCons V E
  | otherwise                         = not (V `elem` (getVars E))
 where
  subConsVar _ (Var _) = True
  subConsCons V1 (Comb ConsCall _ Es) = all (subConsPos V1) Es

getVars E = map (\(_,_,x)->x) (returnVars E)

-- Redundant TRS
--  All variables of l|_i for a symbol F are redundant in r
--  Auxiliary for Theorems
redTRS EF EI FDecls = all (redF EF EI) FDecls
 where
  redF F I (Func FCall _ _ FRule) = F/=FCall || redRule F I FRule
  redRule _ _ (External _) = False
  redRule F I (Rule VarsRule Exp) = RedAllVars
   where
    VarsTriples = collectVarsEra [(F,I,getN I VarsRule)] Exp;
    Vars = map (\(_,_,x)->x) VarsTriples;
    RedAllVars = and (map (\x->x Exp) (map (redVar F I) Vars));

-- First approach to detect redundancy
--  1) all lhs of rule must have a variable at i
--  2) this variable must be (f,i)-redundant in the rhs
redArgsLHSVars F I prog = not (matchedArg ArgI Exp) && redTRS F I FDecls
 where
   Prog _ _ _ FDecls _ _  = prog
   Func _ _ _ (Rule VarsRule Exp) = funcDecl F FDecls
   ArgI = getN I VarsRule

-- Matching variable (variable used in pattern matching)
matchedArg _ (Var _) = False
matchedArg _ (Lit _) = False
matchedArg V (Apply E1 E2) = any (matchedArg V) [E1,E2]
matchedArg V (Constr _ E) = matchedArg V E
matchedArg V (Or E1 E2) = any (matchedArg V) [E1,E2]
matchedArg V (Choice E) = matchedArg V E
matchedArg V (GuardedExpr _ E1 E2) = any (matchedArg V) [E1,E2]
matchedArg V (Comb _ _ Exprs) = any (matchedArg V) Exprs
matchedArg V (Case _ (Var V2) BE) = V==V2 || any (matchedArg V) BE2
 where BE2 = (map (\(Branch _ x)->x) BE)


------------------
-- Second approach

-- Obtain the real set of rules for a function
obtainRules (Func F _ _ (Rule VarsRule Ex)) = obtainExp FCall Ex
 where
  FCall = Comb FuncCall F (map Var VarsRule)
  obtainExp Call Exp
   | is (\(Case _ (Var _) _)->True) Exp = obtainCase Call Exp
   | is (\(Or _ _)->True)     Exp = obtainOr Call Exp
   | otherwise                    = [(Call,Exp)]
  obtainCase Call (Case _ (Var V2) BE) = obtainBranch Call V2 BE
  obtainBranch _    _ [] = []
  obtainBranch Call V ((Branch (Pattern Str Vars) E):xs) =
    obtainExp NewCall E ++ obtainBranch Call V xs
   where NewCall = applySubs (Subs V (Comb ConsCall Str (map Var Vars))) Call
  obtainBranch Call V ((Branch (LPattern Literal) E):xs) =
    obtainExp NewCall E ++ obtainBranch Call V xs
   where NewCall = applySubs (Subs V (Lit Literal)) Call
  obtainOr Call (Or E1 E2) = obtainExp Call E1 ++ obtainExp Call E2

-- Returns a set of triples
--  (rhs of rule1, rhs of rule2, vars of subterms which do not unify)
--  Note that rhs's are just substituted (with common variables)
tripleFI EF EI FDecls = tripleFIAux EI Rules
 where 
  Rules = obtainRules (funcDecl EF FDecls)
  tripleFIAux _ [] = []
  tripleFIAux I ((l,r):xs) = checkAllRules I l r xs
  checkAllRules _ _ _ [] = []
  checkAllRules I l1 r1 ((l2,r2):xs) = 
    if Sols == [] then checkAllRules I l1 r1 xs
                  else process r1 r2 Sols:checkAllRules I l1 r1 xs
   where 
     Sols = findallValues (unifyUpToI I l1 l2)
     process e1 e2 [(sub,vars)] = (e1_,e2_,vars)
      where e1_ = foldFuncs (map applySubs sub) e1
            e2_ = foldFuncs (map applySubs sub) e2

-- If two lhs's unify up to some positions
--   returns a substitution and 
--   the set of variables of subterms at i-th parameter
unifyUpToI I (Comb FuncCall F1 Es1) (Comb FuncCall F2 Es2) 
 | F1 == F2 = (unifyL Es1_ Es2_,returnVars Es1I ++ returnVars Es2I)
 where 
  Es1_ = dropN I Es1
  Es2_ = dropN I Es2
  Es1I = getN I Es1
  Es2I = getN I Es2

-- Unify two terms (note we only consider lhs's)
unify E1 E2
 | is (\(Var _)->True) E1 = unifyVar E1 E2
 | is (\(Var _)->True) E2 = unifyVar E2 E1
 | otherwise              = unifyFF E1 E2
unifyVar (Var V) t = if t==(Var V) then [] else [Subs V t]
unifyFF (Comb _ F1 Es1) (Comb _ F2 Es2) | F1==F2 = unifyL Es1 Es2
unifyFF (Lit Lit1) (Lit Lit2) | Lit1 == Lit2 = []
unifyL (x:xs) (y:ys) = unify x y ++ unifyL xs ys
unifyL [] [] = []

-- Joinable (f,i)-triple
joinTriple F FType DataTypes FuncDecls (r1,r2,vars) = joinable FuncDecls r1e r2e
 where
  joinable FDecls e1 e2 = e1N == e2N
    where e1N = normalizeConfluent FDecls e1
          e2N = normalizeConfluent FDecls e2
  sub = buildSubsts vars F FType DataTypes
  r1e = foldFuncs (map applySubs sub) r1
  r2e = foldFuncs (map applySubs sub) r2

-- Second approach to detect redundancy
--  1) TRS is (f,i)-joinable 
--  2) all lhs of rule must have a variable at i
redArgsLHSPatts EF EI prog = joinTRS EF EI FDecls
 where
  Prog _ _ TypeD FDecls _ _ = prog
  Func _ _ FType (Rule _ _) = funcDecl EF FDecls
  joinTRS F I Fs = redTRS F I Fs && alljoinTriples
   where
    triples = tripleFI F I Fs
    alljoinTriples = and (map (joinTriple F FType TypeD Fs) triples)

---------------------------------------------
-- Algorithm for detecting redundancies
--  application of first and second approach

-- Analyse program
redArgs prog = analyse (args FDecls) [] prog
 where
  Prog _ _ _ FDecls _ _ = prog
  args [] = []
  args ((Func F A _ _):xs) = (argsF 1 F A) ++ (args xs)
  argsF I F A = if (I>A) then [] else ((F,I):argsF (I+1) F A)

-- Analysis
-- First argument is [(F,i)]
-- Second argument is [EFunc F Args]
analyse InitialArgs RArgs program | x_ =:= putStr "." = 
  if ArgsNew == []
  then collect RArgs
  else analyse IArgs RArgsN (eraseProgPartial (collect ArgsNew) program)
 where 
  x_ free
  ArgsNew = analyseAux InitialArgs program
  analyseAux []               _    = []
  analyseAux ((F,I):xs) prog = 
     if redArgsLHSVars F I prog || redArgsLHSPatts F I prog
     then (F,I):analyseAux xs prog
     else analyseAux xs prog
  IArgs = foldFuncs (map delete ArgsNew) InitialArgs
  RArgsN = RArgs ++ ArgsNew
  collect [] = []
  collect ((F,I):xs) = EFunc F Args:collect xs2
   where 
    isF F1 (F2,_) = F1==F2
    AllFI = filter (isF F) ((F,I):xs)
    Args = map snd AllFI
    xs2 = filter (\(F1,_)->F1/=F) ((F,I):xs)

-------------------------
-- Detect orthogonality

orthoProg prog = all orthoFunc FDecls
 where Prog _ _ _ FDecls _ _ = prog

orthoFunc (Func _ _ _ (External _)) = True
orthoFunc (Func F A T (Rule VarsRule Exp)) = orthoRules lhss
 where
  rules = obtainRules (Func F A T (Rule VarsRule Exp))
  lhss = map (\(x,_)->x) rules
  orthoRules [] = True
  orthoRules (x:xs) = orthoRule x xs && orthoRules xs
  orthoRule _ [] = True
  orthoRule L1 (L2:Ls) = findallValues (unify L1 L2) == [] && orthoRule L1 Ls

-----------------------------
-- Detect completely defined

testCDProg prog = all (testCDFunc TypeD) FDecls
 where 
  Prog _ _ TypeD FDecls _ _ = prog

testCDFunc _ (Func _ _ _ (External _)) = True
testCDFunc TypeD (Func FSym Ar FT (Rule VarsRule Exp)) = testCDLEs TypeD Exps
 where
  rules = obtainRules (Func FSym Ar FT (Rule VarsRule Exp))
  lhss = map fst rules
  Exps = map (\(Comb _ _ x)->x) lhss

testCDLEs TypeD LEs = LEs==[] || (test1Arg && all (testCDLEs TypeD) LEs_)
 where
  -- Variables
  VLEs = filter (is (\(Var _:_)->True)) LEs
  anyVar = VLEs /= []
  -- Constructors
  CLEs = filter (is (\(Comb _ _ _:_)->True)) LEs
  consL = nub (map (\(Comb _ F _:_)->F) CLEs)
  consT = if consL==[] then [] else constType (head consL) TypeD
  sameSetCons = eqSets consL consT
  -- Test for expressions in current argument
  test1Arg = anyVar || (sameSetCons && all (testCDLEs TypeD) subEs)
  -- Split of LEs w.r.t. constructors
  splitLEs = map splitLEsF consL
  splitLEsF F1 = filter (is (\(Comb _ F2 _:_)->F1==F2)) LEs
  -- Sub expressions to test 
  subEs = map (map (\(Comb _ _ x:_)->x)) splitLEs
  -- Rest of LEs to test after the subexpressions
  splitVLEs = if consL /= [] && sameSetCons then map (VLEs++) splitLEs 
                                            else VLEs:map (VLEs++) splitLEs
  LEs_ = map (map (\(_:x)->x)) splitVLEs

-- Returns the list of constructor symbols associated to a type
-- The type is identified by a constructor symbol of the type :-)
constType Str TypeD 
  | Str == "[]" || Str == ":"       = ["[]",":"]
  | Str == "True" || Str == "False" = ["True","False"]
  | Str == "(,...,)"                = ["(,...,)"]
  | otherwise                       = consType2 Str TypeD
 where
  consType2 S [] = error ("Constructor symbol "++S++" non existent")
  consType2 S (Type _ _ consD:ts) 
   | S `elem` cL = cL
   | otherwise   = consType2 S ts
   where cL = map (\(Cons x _ _)->x) consD
  
---Tools------------------------------------------------------------------

-- Returns a list (without duplicates) of variables appearing in Exp
--  Note that Exp is a LHS!!!!!
--  including the symbol and argument for type information
returnVars Exp = nub (returnVarsAux "" 1 [] Exp)
 where
  returnVarsAux F N Vars (Var VarIndex) = Vars ++ [(F,N,VarIndex)]
  returnVarsAux _ _ Vars (Lit _) = Vars
  returnVarsAux _ _ Vars (Comb _ FN Exprs) = returnVL FN 1 Vars Exprs
  returnVL _ _ Vars [] = Vars
  returnVL F N Vars (x:xs) = returnVL F (N+1) (returnVarsAux F N Vars x) xs
  --returnVarsAux F N Vars (Apply E1 E2) = returnVL Vars [E1,E2]
  --returnVarsAux F N Vars (Constr _ Expr) = returnVarsAux Vars Expr
  --returnVarsAux F N Vars (Or E1 E2) = returnVL Vars [E1,E2]
  --returnVarsAux F N Vars (Case _ E BE) = getB (returnVarsAux Vars E) BE
  -- where getB Vars [] = Vars
  --       getB Vars ((Branch _ E):xs) = getB (returnVarsAux Vars E) xs
  --returnVarsAux F N Vars (Choice Expr) = returnVarsAux Vars Expr
  --returnVarsAux F N Vars (GuardedExpr _ E1 E2) = returnVL Vars [E1,E2]

-- function for a Exp
getExp I E 
  | is (\(Comb _ _ _)->True) E = getExp2 I E
  | otherwise                  = error "getExp incorrectly call"
 where 
  getExp2 I2 (Comb _ _ Expr) = getN I2 Expr

-- compare two sets
eqSets [] ys = ys==[]
eqSets (x:xs) ys = x `elem` ys && eqSets xs (delete x ys)

-- Find all solutions to a term
findallValues p = findall (\x->p =:= x)

-- end of program
