------------------------------------------------------------------------------
-- Generate an interface description for a Curry module.
-- The interface description contains the type declarations
-- for all functions by this module.
--
-- Michael Hanus, March 2001
-- Modified by Santiago Escobar for 
-- the elimination of redundant arguments (November 2001)
------------------------------------------------------------------------------

module genint(flc2curry) where

import Flat
import List
import Char

-- Obtain a human-readable representation of a program 
-- by using curry tool library "genint"
flc2curry file= 
  do modstring <- genCurryMod fileFLC
     writeFile fileCurry ("--module "++file++" where\n")
     appendFile fileCurry modstring
  where
   fileFLC = file
   fileCurry = file ++ ".curry"

-- show interface on stdout:
showInterface :: String -> IO ()
showInterface progname =
  do intstring <- genInt progname
     putStrLn ("Interface of module \""++progname++"\":\n")
     putStrLn intstring

-- write interface into interface file <prog>.cif:
writeInterface :: String -> IO ()
writeInterface progname =
  do intstring <- genInt progname
     writeFile (progname++".cif")
               ("--Interface of module \""++progname++"\":\n\n"++
                intstring)
     putStrLn ("Interface written into file \""++progname++".cif\"")

genInt :: String -> IO String
genInt progname =
  do (Prog _ imports types functions ops transtable) <- readFlatCurry progname
     return $ concatMap genIntImport imports ++ "\n" ++
              concatMap (genIntOpDecl (translate transtable)) ops ++
              (if null ops then "" else "\n") ++
              concatMap (genIntType (translate transtable)
                                    (map (\(Trans _ en)->en) transtable))
                                    types ++ "\n" ++
              concatMap (genIntFunc (translate transtable)
                                    (map (\(Trans _ en)->en) transtable))
                                    functions ++ "\n"
  where
    translate [] f = f
    translate (Trans name intname : trs) f
        | f==intname  = name
        | otherwise   = translate trs f

-- write import declaration
--genIntImport impmod = "import "++impmod++"\n"
genIntImport impmod = "--import "++impmod++"\n"

-- write operator declaration
genIntOpDecl tf (Op op InfixOp  prec) = "infix "++show prec++" "++tf op++"\n"
genIntOpDecl tf (Op op InfixlOp prec) = "infixl "++show prec++" "++tf op++"\n"
genIntOpDecl tf (Op op InfixrOp prec) = "infixr "++show prec++" "++tf op++"\n"

-- write datatype declaration
genIntType tf expnames (Type tcons tvars constrs) =
  if tcons `elem` expnames
  then "data " ++ tf tcons ++ concatMap (\i->[' ',chr (97+i)]) tvars ++
       " = " ++
       concat (intersperse " | " (map (showExportConsDecl tf expnames) constrs))
       ++ "\n"
  else ""

showExportConsDecl tf expnames (Cons cname _ argtypes) =
  if cname `elem` expnames
  then tf cname ++ concatMap (\t->" "++showCurryType tf True t) argtypes
  else ""

-- write function type declaration
genIntFunc trans expnames (Func fname _ ftype _) =
  if fname `elem` expnames
  then trans fname ++" :: "++ showCurryType trans False ftype ++ "\n"
  else ""

-- pretty printer for types in Curry syntax:
-- second argument is True iff brackets must be written around complex types
showCurryType :: (String -> String) -> Bool -> TypeExpr -> String
showCurryType _ _ (TVar i) = [chr (97+i)]
showCurryType tf nested (FuncType t1 t2) =
   (if nested then "(" else "") ++
     showCurryType tf (isFuncType t1) t1 ++ " -> " ++
     showCurryType tf False t2 ++
   (if nested then ")" else "")
showCurryType tf nested (TCons tc ts)
 | ts==[]  = tf tc
 | tc=="[]" && (head ts == TCons "Char" []) = "String"
 | tc=="[]" = "[" ++ showCurryType tf False (head ts) ++ "]" -- list type
 | take 2 tc=="(,"                                           -- tuple type
   = "(" ++ concat (intersperse "," (map (showCurryType tf False) ts)) ++ ")"
 | otherwise
   = (if nested then "(" else "") ++
       tf tc ++ " " ++
         concat (intersperse " " (map (showCurryType tf True) ts)) ++
     (if nested then ")" else "")

isFuncType (TVar _) = False
isFuncType (FuncType _ _) = True
isFuncType (TCons _ _) = False


---------------------------------------------------------------------------
-- generate a human-readable representation of a Curry module:

-- show representation on stdout:
showCurryMod :: String -> IO ()
showCurryMod progname =
  do modstring <- genCurryMod progname
     putStrLn ("--module "++progname++" where\n")
     putStrLn modstring

genCurryMod :: String -> IO String
genCurryMod progname =
  do (Prog _ imports types functions ops transtable) <- readFlatCurry progname
     return $ concatMap genIntImport imports ++ "\n" ++
              concatMap (genIntOpDecl (translate transtable)) ops ++
              (if null ops then "" else "\n") ++
              concatMap (showCurryDataDecl (translate transtable)) types
              ++ "\n" ++
              concatMap (showCurryFuncDecl (translate transtable)) functions
              ++ "\n-- end of module " ++ progname ++ "\n"
  where
    translate [] f = f
    translate (Trans name intname : trs) f
        | f==intname  = name
        | otherwise   = translate trs f

showCurryDataDecl tf (Type tcons tvars constrs) =
  "data " ++ tf tcons ++ concatMap (\i->[' ',chr (97+i)]) tvars ++
  " = " ++
  concat (intersperse " | " (map (showCurryConsDecl tf) constrs))
  ++ "\n"

showCurryConsDecl tf (Cons cname _ argtypes) =
  tf cname ++ concatMap (\t->" "++showCurryType tf True t) argtypes


-- generate function definitions:
showCurryFuncDecl tf (Func fname _ ftype frule) =
  tf fname ++" :: "++ showCurryType tf False ftype ++ "\n" ++
  showCurryRule tf fname frule

showCurryRule tf fname (External _) = tf fname ++ " external\n\n"
showCurryRule tf fname (Rule lhs rhs) =
  --showCurryRuleAsCase tf fname (Rule lhs rhs)
  showCurryRuleAsPatterns tf fname (Rule lhs rhs)

-- format rule as case expression:
showCurryRuleAsCase tf fname (Rule lhs rhs) =
   tf fname ++ " " ++ showCurryElems showCurryVar lhs ++
   " = " ++ showCurryExpr tf False 0 rhs ++ "\n\n"

-- format rule as set of pattern matching rules:
showCurryRuleAsPatterns tf fname (Rule lhs rhs) =
   showEvalAnnot (getFlexRigid rhs) ++
   concatMap (\(l,r)->showCurryPatternRule tf l r) patternrules
   ++ "\n"
 where
   showEvalAnnot ConflictFR = tf fname ++ " eval rigid&flex -- CONFLICTING!!\n"
   showEvalAnnot UnknownFR = ""
   showEvalAnnot KnownRigid = tf fname ++ " eval rigid\n"
   showEvalAnnot KnownFlex  = "" --tf fname ++ " eval flex\n"

   patternrules = rule2equations (shallowPattern2Expr fname lhs) rhs

showCurryPatternRule tf l r =
   showCurryExpr tf False 0 l ++
   (if isGuardedExpr r
    then " | " ++ showCurryCondRule r
    else " = " ++ showCurryExpr tf False 2 r) ++ "\n"
 where
   showCurryCondRule (GuardedExpr [] e1 e2) =
     showCurryExpr tf False 2 e1 ++ " = " ++ showCurryExpr tf False 4 e2
   showCurryCondRule (GuardedExpr (x:xs) e1 e2) =
     showCurryExpr tf False 2 e1 ++ " = " ++ showCurryExpr tf False 4 e2 ++
     " where " ++ concat (intersperse "," (map showCurryVar (x:xs))) ++ " free"


-- compute the rigid/flex status of an expression, i.e., check all
-- cases in this expression (or fail, if the expression has rigid as well as
-- flex cases):

-- datatype for computed flex/rigid result:
data FlexRigidResult = UnknownFR | ConflictFR | KnownFlex | KnownRigid

getFlexRigid :: Expr -> FlexRigidResult
getFlexRigid (Var _) = UnknownFR
getFlexRigid (Lit _) = UnknownFR
getFlexRigid (Comb _ _ args) =
   foldr joinCaseTypes UnknownFR (map getFlexRigid args)
getFlexRigid (Apply e1 e2) =
   joinCaseTypes (getFlexRigid e1) (getFlexRigid e2)
getFlexRigid (Constr _ e) = getFlexRigid e
getFlexRigid (Or e1 e2) =
   joinCaseTypes (getFlexRigid e1) (getFlexRigid e2)
getFlexRigid (Case ctype e bs) =
   foldr joinCaseTypes (if ctype==Flex then KnownFlex else KnownRigid)
         (map getFlexRigid (e : map (\(Branch _ be)->be) bs))
getFlexRigid (Choice e) = getFlexRigid e
getFlexRigid (GuardedExpr _ e1 e2) =
   joinCaseTypes (getFlexRigid e1) (getFlexRigid e2)

joinCaseTypes ConflictFR UnknownFR = ConflictFR
joinCaseTypes ConflictFR KnownFlex = ConflictFR
joinCaseTypes ConflictFR KnownRigid = ConflictFR
joinCaseTypes UnknownFR ConflictFR = ConflictFR
joinCaseTypes KnownFlex ConflictFR = ConflictFR
joinCaseTypes KnownRigid ConflictFR = ConflictFR
joinCaseTypes UnknownFR UnknownFR = UnknownFR
joinCaseTypes UnknownFR KnownFlex = KnownFlex
joinCaseTypes UnknownFR KnownRigid = KnownRigid
joinCaseTypes KnownFlex UnknownFR = KnownFlex
joinCaseTypes KnownFlex KnownFlex = KnownFlex
joinCaseTypes KnownFlex KnownRigid = ConflictFR
joinCaseTypes KnownRigid UnknownFR = KnownRigid
joinCaseTypes KnownRigid KnownFlex = ConflictFR
joinCaseTypes KnownRigid KnownRigid = KnownRigid


-- transform a rule consisting of a left- and a right-hand side
-- (represented as expressions) into a set of pattern matching rules:
rule2equations :: Expr -> Expr -> [(Expr,Expr)]
rule2equations lhs (Or e1 e2) =
   rule2equations lhs e1 ++ rule2equations lhs e2
rule2equations lhs (Case ctype e bs) =
   if isVarExpr e then let Var i = e  in  caseIntoLhs lhs i bs
                  else [(lhs,Case ctype e bs)]
rule2equations lhs (Var i) = [(lhs,Var i)]
rule2equations lhs (Lit l) = [(lhs,Lit l)]
rule2equations lhs (Comb ct name args) = [(lhs,Comb ct name args)]
rule2equations lhs (Apply e1 e2) = [(lhs,Apply e1 e2)]
rule2equations lhs (Constr vs e) = [(lhs,Constr vs e)]
rule2equations lhs (Choice e) = [(lhs,Choice e)]
rule2equations lhs (GuardedExpr vs e1 e2) = [(lhs,GuardedExpr vs e1 e2)]

caseIntoLhs _ _ [] = []
caseIntoLhs lhs vi (Branch (Pattern c vs) e : bs) =
  rule2equations (substitute [vi] [shallowPattern2Expr c vs] lhs) e
  ++ caseIntoLhs lhs vi bs
caseIntoLhs lhs vi (Branch (LPattern lit) e : bs) =
  rule2equations (substitute [vi] [Lit lit] lhs) e
  ++ caseIntoLhs lhs vi bs

shallowPattern2Expr name vars = Comb ConsCall name (map (\i->Var i) vars)


-- (substitute vars exps expr) = expr[vars/exps]
-- i.e., replace all occurrences of vars by corresponding exps in the
-- expression expr
substitute vars exps expr = substituteAll vars exps 0 expr

-- (substituteAll vars exps base expr):
-- substitute all occurrences of variables by corresonding expressions:
-- * substitute all occurrences of var_i by exp_i in expr
--   (if vars=[var_1,...,var_n] and exps=[exp_1,...,exp_n])
-- * substitute all other variables (Var j) by (Var (base+j))
--
-- here we assume that the new variables in guards and case patterns
-- do not occur in the list "vars" of replaced variables!

substituteAll :: [Int] -> [Expr] -> Int -> Expr -> Expr
substituteAll vars exps b (Var i) = replaceVar vars exps i
  where replaceVar [] [] var = Var (b+var)
        replaceVar (v:vs) (e:es) var = if v==var then e
                                                 else replaceVar vs es var
substituteAll _  _  _ (Lit l) = Lit l
substituteAll vs es b (Comb combtype c exps) =
                 Comb combtype c (map (substituteAll vs es b) exps)
substituteAll vs es b (Apply e1 e2) =
                 Apply (substituteAll vs es b e1) (substituteAll vs es b e2)
substituteAll vs es b (Constr vars e) =
                 Constr (map (+b) vars) (substituteAll vs es b e)
substituteAll vs es b (Or e1 e2) =
                 Or (substituteAll vs es b e1) (substituteAll vs es b e2)
substituteAll vs es b (Case ctype e cases) =
   Case ctype (substituteAll vs es b e) (map (substituteAllCase vs es b) cases)
substituteAll vs es b (GuardedExpr vars c e) =
   GuardedExpr (map (+b) vars) (substituteAll vs es b c)
                               (substituteAll vs es b e)
substituteAll vs es b (Choice e) = Choice (substituteAll vs es b e)

substituteAllCase vs es b (Branch (Pattern l pvs) e) =
                 Branch (Pattern l (map (+b) pvs)) (substituteAll vs es b e)
substituteAllCase vs es b (Branch (LPattern l) e) =
                 Branch (LPattern l) (substituteAll vs es b e)



-- (almost) Curry printer for expressions
-- first argument: translation function for names
-- second argument is True iff brackets must be written around complex types
-- third argument: indentation (for Case/Or expressions)
showCurryExpr :: (String -> String) -> Bool -> Int -> Expr -> String

showCurryExpr _ _ _ (Var n) = showCurryVar n

showCurryExpr _ _ _ (Lit l) = showCurryLit l

showCurryExpr tf _ _ (Comb _ cf []) = tf cf
showCurryExpr tf nested b (Comb _ cf [e]) =
  maybeShowBrackets nested (tf cf ++ " " ++ showCurryExpr tf True b e)
showCurryExpr tf nested b (Comb ct cf [e1,e2]) =
  if isAlpha (head (tf cf))
  then maybeShowBrackets nested
            (tf cf ++ " " ++ showCurryElems (showCurryExpr tf True b) [e1,e2])
  else
   if isFiniteList (Comb ct cf [e1,e2])
   then "[" ++
         concat
            (intersperse "," (showCurryFiniteList tf b (Comb ct cf [e1,e2])))
        ++ "]"
   else
    if cf=="(,)" -- pair constructor?
    then "(" ++ showCurryExpr tf False b e1 ++ "," ++
                showCurryExpr tf False b e2 ++ ")"
    else maybeShowBrackets nested
              (showCurryExpr tf True b e1 ++ " " ++ tf cf ++ " " ++
               showCurryExpr tf True b e2 )
showCurryExpr tf nested b (Comb _ cf (e1:e2:e3:es)) =
  if cf=="if_then_else" && es==[]
  then maybeShowBrackets nested
        ("\n" ++
         blanks b ++ "if "   ++ showCurryExpr tf False (b+2) e1 ++ "\n" ++
         blanks b ++ "then " ++ showCurryExpr tf False (b+2) e2 ++ "\n" ++
         blanks b ++ "else " ++ showCurryExpr tf False (b+2) e3)
  else
   if take 2 cf == "(,"  -- tuple constructor?
   then "(" ++
         concat
            (intersperse "," (map (showCurryExpr tf False b) (e1:e2:e3:es)))
        ++ ")"
   else maybeShowBrackets nested
       (tf cf ++ " " ++ showCurryElems (showCurryExpr tf True b) (e1:e2:e3:es))

showCurryExpr tf nested b (Apply e1 e2) =
  maybeShowBrackets nested
       (showCurryExpr tf True b e1 ++ " " ++ showCurryExpr tf True b e2)

showCurryExpr tf nested b (Constr [] e) = showCurryExpr tf nested b e

showCurryExpr tf nested b (Constr (x:xs) e) =
  maybeShowBrackets nested
    ("let " ++ concat (intersperse "," (map showCurryVar (x:xs))) ++
     " free in " ++ showCurryExpr tf False b e)

showCurryExpr tf nested b (Or e1 e2) =
  maybeShowBrackets nested
    ("Or " ++ showCurryExpr tf True b e1 ++ " " ++ showCurryExpr tf True b e2)

showCurryExpr tf nested b (Case ctype e cs) =
  maybeShowBrackets nested
    ((if ctype==Rigid then "Case " else "FCase ") ++
     showCurryExpr tf True b e ++ " of\n " ++
     showCurryElems (showCurryCase tf (b+2)) cs ++ blanks b)

showCurryExpr tf nested b (GuardedExpr [] e1 e2) =
  maybeShowBrackets nested
    (showCurryExpr tf False b e1 ++ " => " ++ showCurryExpr tf False b e2)
showCurryExpr tf nested b (GuardedExpr (x:xs) e1 e2) =
  maybeShowBrackets nested
   (showCurryExpr tf False b e1 ++ " => " ++ showCurryExpr tf False b e2 ++ 
    " where " ++ concat (intersperse "," (map showCurryVar (x:xs))) ++ " free")
 
showCurryExpr tf nested b (Choice e) =
  maybeShowBrackets nested ("Choice " ++ showCurryExpr tf True b e)


showCurryVar i = "v" ++ show i

showCurryLit (Intc   i) = show i
showCurryLit (Floatc f) = show f
showCurryLit (Charc  c) = "'" ++ [c] ++ "'"

showCurryCase tf b (Branch (Pattern l xs) e) =
  blanks b ++ tf l ++ " " ++ showCurryElems showCurryVar xs
                   ++ " -> " ++ showCurryExpr tf False b e ++ "\n"
showCurryCase tf b (Branch (LPattern l) e) =
  blanks b ++ showCurryLit l ++ " "
           ++ " -> " ++ showCurryExpr tf False b e ++ "\n"

showCurryFiniteList _ _ (Comb _ "[]" []) = []
showCurryFiniteList tf b (Comb _ ":" [e1,e2]) =
  showCurryExpr tf False b e1 : showCurryFiniteList tf b e2

-- format a finite list of elements (with a format function for list elems):
showCurryElems :: (a->String) -> [a] -> String
showCurryElems format elems =
   concat (intersperse " " (map format elems))

maybeShowBrackets nested s =
   (if nested then "(" else "") ++ s ++ (if nested then ")" else "")

blanks b = take b (repeat ' ')

---------------------------------------------------------------------------
-- Auxiliary functions
---------------------------------------------------------------------------

-- is the expression a finite list (with an empty list at the end)?
isFiniteList :: Expr -> Bool
isFiniteList (Var _) = False
isFiniteList (Lit _) = False
isFiniteList (Comb _ name args)
  | name=="[]" && args==[] = True
  | name==":"  && length args == 2 = isFiniteList (args!!1)
  | otherwise = False
isFiniteList (Apply _ _) = False
isFiniteList (Constr _ _) = False
isFiniteList (Or _ _) = False
isFiniteList (Case _ _ _) = False
isFiniteList (Choice _) = False
isFiniteList (GuardedExpr _ _ _) = False

-- is the expression a guarded expressions?
isGuardedExpr :: Expr -> Bool
isGuardedExpr (Var _) = False
isGuardedExpr (Lit _) = False
isGuardedExpr (Comb _ _ _) = False
isGuardedExpr (Apply _ _) = False
isGuardedExpr (Constr _ _) = False
isGuardedExpr (Or _ _) = False
isGuardedExpr (Case _ _ _) = False
isGuardedExpr (Choice _) = False
isGuardedExpr (GuardedExpr _ _ _) = True

-- is the expression a variable?
isVarExpr :: Expr -> Bool
isVarExpr (Var _) = True
isVarExpr (Lit _) = False
isVarExpr (Comb _ _ _) = False
isVarExpr (Apply _ _) = False
isVarExpr (Constr _ _) = False
isVarExpr (Or _ _) = False
isVarExpr (Case _ _ _) = False
isVarExpr (Choice _) = False
isVarExpr (GuardedExpr _ _ _) = False


-- Examples:
-- showInterface "genint"
-- showInterface "../examples/nats"
-- showInterface "../examples/maxtree"
-- showInterface "../lib/Flat"
-- writeInterface "genint"
-- showCurryMod "genint"
-- showCurryMod "../examples/rev"


-- end of program

-- end of program
