------------------------------------------------------------------------
--      Meta Interpreter of standard Curry Semantics
--     by Mara Alpuente, Santiago Escobar and Salvador Lucas
--    Version of 16/11/2001 (Universidad Politcnica de Valencia)
--   **Thanks to Germn Vidal and Elvira Albert for their help**
------------------------------------------------------------------------

------------------------------------------------------------------
------- Meta Interpreter of standard Curry semantics
------------------------------------------------------------------

import Flat

-----------------------
---------General stuff

-- data type for substitutions:
data Subst = Sub [Int] [Expr] | FSub

--data PEvalState = PEState [Expr] [FuncDecl] Int Int
data ProState = PState [FuncDecl] Int

getFuncDecls     (PState fds _) = fds
getRenamingIndex (PState _ i ) = i

-- update state with new index for renamed variables:
updateRenamingIndex :: ProState -> Int -> ProState
updateRenamingIndex (PState fds _) i = PState fds i

-- compute arity of a function symbol:
getArity :: ProState -> String -> Int
getArity pst f = getArityF (getFuncDecls pst) f

getArityF :: [FuncDecl] -> String -> Int
getArityF funs f
 | f=="+"    = 2
 | f=="-"    = 2
 | f=="*"    = 2
 | f=="<"    = 2
 | f==">"    = 2
 | f=="<="   = 2
 | f==">="   = 2
 | f=="&"    = 2
 | f=="&&"   = 2
 | f=="=:="  = 2
 | f=="=="   = 2
 | f=="if_then_else" = 3
 | f=="map"          = 2
 | f=="foldl"        = 3
 | f=="foldr"        = 3
 | f=="filter"       = 2
 | f=="iterate"      = 2
 | otherwise = findArity funs
 where
    findArity [] = error ("INTERNAL ERROR: cannot find arity of "++f)
    findArity (Func fn fa _ _ : fds) =
        if fn==f then fa else findArity fds

----------------------
--- Different Calls

-- One Curry semantics step
evalStep :: [FuncDecl] -> (Expr,Subst) -> (Expr,Subst)
evalStep funs (e,s) = (ne,ns)
 where 
  (ne,ns) = evalc (PState funs (maxVarIndex e)) (e,s)

-- Normal forms Curry Semantics
evalNorm :: [FuncDecl] -> (Expr,Subst) -> (Expr,Subst)
evalNorm funs (e,s) = if e==ne then (e,s) else evalNorm funs (ne,ns)
   where (ne,ns) = evalStep funs (e,s)

-- Normalize by rewriting
normalize :: [FuncDecl] -> Expr -> Expr
normalize funs E = fst (normalizeAux funs (E,Sub [] []))
 where 
  normalizeAux funsA (e,s) =
     if e==ne || ns/=(Sub [] []) || (isSuspended ne) || (isInstantiation ne)
     then (e,s) else normalizeAux funsA (ne,ns)
   where 
     (ne,ns) = evalStep funsA (e,s)

isInstantiation :: Expr -> Bool
isInstantiation e = first (isInstantiationAux e)
 where
  isInstantiationAux (Case Flex (Var _) _) = True
  isInstantiationAux (Comb _ _ es) = orL (map isInstantiationAux es)
  isInstantiationAux _ = False
  orL es = foldr (||) False es

-- Normalize in a confluent system
normalizeConfluent funs e = first (normalize funs e)

----------------------------------------------------------------------------
--evalc implements a one standard curry semantics step

evalc :: ProState -> (Expr,Subst) -> (Expr,Subst)

----------------------------------------------------------------------------
--RLNT calculus:
--
-- HNF, Case of Case, Case Function, Case Select, Case Guess, Function Eval
----------------------------------------------------------------------------

----------------------------------------------------------------------------
-- HNF:

evalc _ (Var v,s) = (Var v,s)
evalc _ (Lit l,s) = (Lit l,s)
evalc pst (Comb ConsCall c es, s) = evalc_cons pst c es [] s

evalc_cons _ c [] es s = (Comb ConsCall c es, s)
evalc_cons pst c (e:es) zs s = 
  if ne==e 
     then evalc_cons pst c es (zs++[e]) s
     else (Comb ConsCall c (nzs++[ne]++nes), ns)
  where (ne,ns) = evalc pst (e,s)
        Sub vars exps = ns
        nzs = map (substitute vars exps) zs
        nes = map (substitute vars exps) es

------------------------------------------------------------------------------
-- Case Select:

evalc _ (Case _ (Comb ConsCall f es) ces, s) = (matchBranch ces f es, s)

-- match a branch (for Case Select):
matchBranch :: [BranchExpr] -> String -> [Expr] -> Expr
matchBranch [] _ _ = (Comb FuncCall "failed" [])
matchBranch (Branch (Pattern p vars) e : ces) c es =
  if p==c then substitute vars es e else matchBranch ces c es

--this is for Case's with a constant argument:
evalc _ (Case _ (Lit l) ces, s) = (matchBranchLit ces l, s)

-- match a branch with Literals:
matchBranchLit :: [BranchExpr] -> Literal -> Expr
matchBranchLit [] _ = (Comb FuncCall "failed" [])
matchBranchLit (Branch (LPattern p) e : ces) c =
  if p==c then e else matchBranchLit ces c

------------------------------------------------------------------------------
-- Case Guess:

evalc pst (Case ctype (Var v) ces, s) =
  if ctype==Flex
        then evalc_cguess pst v ces s
        else (Case ctype (Var v) ces, s)

evalc_cguess _ v ces s =
  if length ces == 1
     then (propBindingOneNew2 v ces s)
     else (propBindingOneNew v ces s)

--propagate matchings to the corresponding branches:
propBindingOne :: Int -> [BranchExpr] -> Expr
propBindingOne v pes = propBinding v (mem pes)

mem (a:_) = a
mem (_:b) = mem b

propBindingOneNew2 v pes s = (vce, composeSubs s ns)
  where ce = memNew2 pes
        (vce,ns) = propBinding2 v ce

memNew2 (a:_) = a
memNew2 (_:b) = memNew b

---
-- composition of substitutions:

composeSubs :: Subst -> Subst -> Subst
composeSubs (Sub vars1 exps1) (Sub vars2 exps2) = 
  (Sub (vars1 ++ vars2) (newExps ++ exps2))
  where newExps = (map (substitute vars2 exps2) exps1)
composeSubs FSub (Sub _ _) = FSub
composeSubs (Sub _ _) FSub = FSub
composeSubs FSub FSub = FSub

----------------
propBindingOneNew v pes s = (vce, composeSubs s ns)
  where ce = memNew pes
        (vce,ns) = propBinding2 v ce

memNew (a:_) = a
memNew (_:b) = memNew b

propBinding2 v (Branch (Pattern c vs) e) = 
   (substitute [v] [(Comb ConsCall c (index2vars vs))] e, 
    Sub [v] [(Comb ConsCall c (index2vars vs))])
propBinding2 v (Branch (LPattern (Intc p)) e) = 
   (substitute [v] [(Lit (Intc p))] e,
    Sub [v] [(Lit (Intc p))])
propBinding2 v (Branch (LPattern (Floatc p)) e) = 
   (substitute [v] [(Lit (Floatc p))] e,
    Sub [v] [(Lit (Floatc p))])
propBinding2 v (Branch (LPattern (Charc p)) e) = 
   (substitute [v] [(Lit (Charc p))] e,
    Sub [v] [(Lit (Charc p))])



--propagate matchings to the corresponding branches:
propBinding :: Int -> BranchExpr -> Expr

propBinding v (Branch (Pattern c vs) e) = 
   substitute [v] [(Comb ConsCall c (index2vars vs))] e
propBinding v (Branch (LPattern (Intc p)) e) = 
   substitute [v] [(Lit (Intc p))] e
propBinding v (Branch (LPattern (Floatc p)) e) = 
   substitute [v] [(Lit (Floatc p))] e
propBinding v (Branch (LPattern (Charc p)) e) = 
   substitute [v] [(Lit (Charc p))] e

-- transform variable indices into variable expressions and vice versa:
index2vars :: [VarIndex] -> [Expr]
index2vars = map (\i->Var i)

vars2index :: [Expr] -> [VarIndex]
vars2index = map (\(Var i)->i)


---------------------------------------------------------------------------
--Case Eval:

evalc pst (Case ctype (Comb FuncCall f es) ces, s) =
  if ne==(Comb FuncCall f es) 
     then (Case ctype (Comb FuncCall f es) ces, s)
     else (Case ctype ne nces, ns)
  where (ne, ns) = evalc pst (Comb FuncCall f es, s)
        Sub vars exps = ns
        nces = map (substituteBranch vars exps) ces

evalc pst (Case ctype (Case ctype2 e2 ces2) ces, s) =
  if ne==(Case ctype2 e2 ces2)
     then (Case ctype (Case ctype2 e2 ces2) ces, s)
     else (Case ctype ne nces, ns)
  where (ne, ns) = evalc pst (Case ctype2 e2 ces2, s)
        Sub vars exps = ns
        nces = map (substituteBranch vars exps) ces

substituteBranch vars exps (Branch p e) = Branch p (substitute vars exps e)


---------------------------------------------------------------------------
--Function Eval:

evalc pst (Comb FuncCall f es, s) = 
  if take 3 (reverse f)=="CCS" 
        then evalc_scc pst f es s
        else if definedFunc pst f
             then evalc_fun pst f es s
             else evalc_builtin pst f es s

extractExpr [_,a] = a
extractCC [Lit (Intc b),_] = b
extractCC2 [cc,_] = cc
extractCC3 (Lit (Intc b)) = b

evalc_fun pst f es s = (newexp, s)
  where (_,newexp) = matchRHS pst f es

evalc_scc pst f es s 
  | isVar e        = (e, s)
  | isLitInt e     = (e, s)
  | isConstant e   = (e, s)
  | isSCC e        = (e, s)
  | isConsCall e   = mixSCC pst f e s
  | otherwise      = (Comb FuncCall f [ne], ns)  --TODO!!!!!!
  where e = extractExpr es
        (ne,ns) = evalc pst (e, s)
        
mixSCC _ _ (Comb ConsCall c es) s =(Comb ConsCall c es,s)

--check whether an expression is a variable:
isSCC :: Expr -> Bool
isSCC x = first (isSCCAux x)
isSCCAux:: Expr -> Bool
isSCCAux (Comb FuncCall f _) = take 3 (reverse f)=="CCS"
isSCCAux _ = False

---------------------------------------------------------------------------
-- CALL UNFOLDING:

-- match a right-hand side of a given function:
matchRHS pst f es = getMatchedRHS pst (getFuncDecls pst) f es 


getMatchedRHS pst (Func fname _ _ funrule : fds) name es =
   if fname==name then getMatchedRHS_aux pst funrule es 
                  else getMatchedRHS pst fds name es 


getMatchedRHS_aux pst (Rule vars rhs) es  =
  let curr_ri = getRenamingIndex pst
      maxindex = max (maxList vars) (maxVarIndex rhs)
  in
   (updateRenamingIndex pst (curr_ri+maxindex+1),
    substituteAll vars es curr_ri rhs)
   -- here we assume that no more than maxExpVars variables occur in each
   -- unfolded expression to produce a safe renaming.

----------------------------------------------------------------------------
-- is f a defined function?
definedFunc :: ProState -> String -> Bool
definedFunc pst f = any (hasName f) (getFuncDecls pst)
definedFunc2 :: [FuncDecl] -> String -> Bool
definedFunc2 funs f = any (hasName f) funs

hasName :: String -> FuncDecl -> Bool
hasName name (Func fname _ _ _) = fname==name





------------------------------------------------------------------------------
-- Pretty print 
------------------------------------------------------------------------------

-- pretty printer for expressions (first argument: indentation):
ppExpr _ (Var n) = ppVar n
ppExpr _ (Lit l) = ppLit l
ppExpr b (Comb _ cf es) =
  "(" ++ cf ++ " " ++ ppList (ppExpr b) es ++ ")"
ppExpr b (Apply e1 e2) =
  "(" ++ ppExpr b e1 ++ " " ++ ppExpr b e2 ++ ")"
ppExpr b (Constr xs e) =
  "(Constr " ++ ppFormatList ppVar xs ++ (ppExpr b) e ++ ")"
ppExpr b (Or e1 e2) =
  "(Or " ++ ppExpr b e1 ++ " " ++ ppExpr b e2 ++ ")"
ppExpr b (Case Rigid e cs) =
  "(Case " ++ ppExpr b e ++ " of\n " ++ ppList (ppCase (b+2)) cs
  ++ blanks b ++ ")"
ppExpr b (Case Flex e cs) =
  "(FCase " ++ ppExpr b e ++ " of\n " ++ ppList (ppCase (b+2)) cs
  ++ blanks b ++ ")"
ppExpr b (GuardedExpr xs e1 e2) =
  "(GuardedExpr " ++ ppFormatList ppVar xs
                  ++ ppExpr b e1 ++ " " ++ ppExpr b e2 ++ ")"
ppExpr b (Choice e) = "(Choice " ++ ppExpr b e ++ ")"

ppVar i = "v" ++ show i

ppLit (Intc   i) = show i
ppLit (Floatc f) = show f
ppLit (Charc  c) = "'" ++ show (ord c) ++ "'"

ppCase b (Branch (Pattern l xs) e) = blanks b ++ "(" ++ l ++ " "
                                     ++ ppList ppVar xs
                                     ++ " -> " ++ ppExpr b e ++ ")\n"
ppCase b (Branch (LPattern l) e) = blanks b ++ "(" ++ ppLit l ++ " "
                                   ++ " -> " ++ ppExpr b e ++ ")\n"

-- REMOVED (included in genint.curry)
--blanks b = take b (repeat ' ')

-- format a finite list of elements (with a format function for list elems):
ppFormatList :: (a->String) -> [a] -> String
ppFormatList format elems = " [" ++ ppElems elems ++ "] "
 where
   ppElems [] = ""
   ppElems [x] = format x
   ppElems (x1:x2:xs) = format x1 ++ "," ++ ppElems (x2:xs)

ppList format elems = ppListElems elems
  where
    ppListElems [] = ""
    ppListElems [x] = format x
    ppListElems (x1:x2:xs) = format x1 ++ " " ++ ppListElems (x2:xs)

------------------------------------------------------------------------------
-- handling variables in expressions:

-- get the maximum index of all variables in an expression:
-- (or -1 if there is no variable)
maxVarIndex :: Expr -> Int
maxVarIndex (Var v)         = v
maxVarIndex (Lit _)         = -1
maxVarIndex (Comb _ _ exps) = maxList (map maxVarIndex exps)
maxVarIndex (Apply e1 e2)   = max (maxVarIndex e1) (maxVarIndex e2)
maxVarIndex (Constr vars e) = max (maxList vars) (maxVarIndex e)
maxVarIndex (Or e1 e2)      = max (maxVarIndex e1) (maxVarIndex e2)
maxVarIndex (Case _ ce cs)   = max (maxVarIndex ce) (maxList (map maxCase cs))
  where maxCase (Branch (Pattern _ pvs) e) = max (maxList pvs) (maxVarIndex e)
        maxCase (Branch (LPattern _)   e) = maxVarIndex e
maxVarIndex (GuardedExpr vars c e) =
                   max (max (maxList vars) (maxVarIndex c)) (maxVarIndex e)
maxVarIndex (Choice e)      = maxVarIndex e

-- maximum of two numbers:
max i j = if i<=j then j else i

-- maximum of a list of naturals (-1 if list is empty):
maxList xs = foldr max (-1) xs


----------------------------------------------------------------------------
--Extension of the RLNT calculus:
--
-- partcall, apply, constr, or, guardedexpr, choice
----------------------------------------------------------------------------



----------------------------------------------------------------------------
--Partcall:

evalc _ (Comb PartCall c es, s) = (Comb PartCall c es, s)


----------------------------------------------------------------------------
--Or expression:

evalc pst (Or e1 _, s) = evalc pst (e1, s) 
evalc pst (Or _ e2, s) = evalc pst (e2, s)


----------------------------------------------------------------------------
--Choice:

--evalc _ _ (Choice e) = Choice e

----------------------------------------------------------------------------
--These rules complements Case-of-case, Case Select, Case Fun, and Case Guess:
--
evalc pst (Case ctype (Apply e1 e2) ces, s) = 
  if ne==(Apply e1 e2)
     then (Case ctype (Apply e1 e2) ces, s)
     else (Case ctype ne nces, ns)
  where (ne, ns) = evalc pst (Apply e1 e2, s)
        Sub vars exps = ns
        nces = map (substituteBranch vars exps) ces

evalc pst (Case ctype (Constr vars es) ces, s) = 
  if ne==(Constr vars es)
     then (Case ctype (Constr vars es) ces, s) 
     else (Case ctype ne nces, ns)
  where (ne, ns) = evalc pst (Constr vars es, s)
        Sub vars2 exps = ns
        nces = map (substituteBranch vars2 exps) ces

evalc pst (Case ctype (Or e1 e2) ces, s) = 
  if ne==(Or e1 e2)
     then (Case ctype (Or e1 e2) ces, s)
     else (Case ctype ne nces, ns)
  where (ne, ns) = evalc pst (Or e1 e2, s)
        Sub vars exps = ns
        nces = map (substituteBranch vars exps) ces

evalc pst (Case ctype (GuardedExpr vars gc e) ces, s) =
  if ne==(GuardedExpr vars gc e)
     then (Case ctype (GuardedExpr vars gc e) ces, s)
     else (Case ctype ne nces, ns)
  where (ne, ns) = evalc pst (GuardedExpr vars gc e, s)
        Sub vars2 exps = ns
        nces = map (substituteBranch vars2 exps) ces

--evalc _ _ (Case ctype (Choice e) ces) = Case ctype (Choice e) ces


----------------------------------------------------------------------------
--Apply:

evalc pst (Apply e1 e2, s) = 
  if (isPartCall e1) then evalc_apply pst e1 e2 s
                     else evalc_applyAux pst e1 e2 s

evalc_applyAux pst e e2 s = 
  if ne==e 
     then (Apply e e2, s)
     else (Apply ne ne2, ns)
  where (ne, ns) = evalc pst (e,s)
        Sub vars exps = ns
        ne2 = substitute vars exps e2

isPartCall :: Expr -> Bool
isPartCall x = first (isPartCallAux x)
isPartCallAux :: Expr -> Bool
isPartCallAux (Comb PartCall _ _) = True
isPartCallAux _ = False

evalc_apply pst (Comb PartCall f es) e2 s =
  if (getArity pst f)==(length es + 1)
  then evalc pst (Comb FuncCall f (es++[e2]), s)
  else evalc pst (Comb PartCall f (es++[e2]), s)

-----------------------------------------------------------------------------
--Guarded expressions:

evalc pst (GuardedExpr vars gc e, s) =
  if gc==(Comb FuncCall "success" []) || gc==(Comb ConsCall "True" [])
     then evalc pst (e,s)
     else evalc_guarded pst vars gc e s

evalc_guarded pst vars gc e s = 
  if ngc==gc
     then (GuardedExpr vars gc e, s)
     else (GuardedExpr vars ngc ne, ns)
  where (ngc, ns) = evalc pst (gc, s)
        Sub vars2 exps = ns
        ne = substitute vars2 exps e

-----------------------------------------------------------------------------
--Constraints:

evalc pst (Constr vars e, s)
  | e==(Comb FuncCall "success" [])  = (Comb FuncCall "success" [],s)
  | e==(Comb ConsCall "True" [])     = (Comb FuncCall "success" [],s)
  | otherwise                        = evalc_constr pst vars e s

evalc_constr pst vars e s = 
  if ne==e
     then (Constr vars e, s)
     else (Constr vars ne, ns)
  where (ne, ns) = evalc pst (e,s)

-----------------------------------------------------------------------------
--Built-in's:

strict x | x==x = x

evalc_builtin :: ProState -> String -> [Expr] -> Subst -> (Expr,Subst)

evalc_builtin pst f es s
  | f=="=="   = evalc_builtinEQ pst f es s
  | f=="=:="  = evalc_builtinEQ2 pst f es s
  | f=="if_then_else" = evalc_builtinITE pst es s
  | f=="&&"   = evalc_builtinAND pst es s
  | f=="&"    = evalc_builtinCAND pst es s
  | f=="*"    = evalc_builtinARITH pst f es s
  | f=="+"    = evalc_builtinARITH pst f es s
  | f=="-"    = evalc_builtinARITH pst f es s
  | f=="<"    = evalc_builtinARITH pst f es s
  | f==">"    = evalc_builtinARITH pst f es s
  | f=="<="   = evalc_builtinARITH pst f es s
  | f==">="   = evalc_builtinARITH pst f es s
  --| f=="map"  = evalc_builtinMAP pst f es 
  --| f=="foldl"   = evalc_builtinFOLDL pst f es
  --| f=="foldr"   = evalc_builtinFOLDR pst f es
  --| f=="filter"  = evalc_builtinFILTER pst f es
  --| f=="iterate" = evalc_builtinITERATE pst f es
  | True        = (Comb FuncCall f es, s)
     
---------------------------------
-- STRICT EQUALITY:

evalc_builtinEQ :: ProState -> String -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinEQ pst f [e1,e2] s
  | (isVar e1) || (isVar e2)            = (Comb FuncCall f [e1,e2], s)
  | (isConsCall e1) && (isConsCall e2)  = evalc_builtinEQaux pst f [e1,e2] s
  | (isLitInt e1) && (isLitInt e2)      = evalc_builtinEQlit f e1 e2 s
  | (isLitInt e1) || (isConsCall e1)    = evalc_EQAux2 pst f e1 e2 s
  | otherwise                           = evalc_EQAux pst f e1 e2 s

evalc_EQAux pst f e1 e2 s = 
  if ne1==e1 then (Comb FuncCall f [ne1,ne2], ns)
             else (Comb FuncCall f [ne1,ne2], ns)
  where (ne1, ns) = evalc pst (e1,s)
        Sub vars exps = ns
        ne2 = substitute vars exps e2

evalc_EQAux2 pst f e1 e2 s = 
  if ne2==e2
     then (Comb FuncCall f [e1,e2], s)
     else (Comb FuncCall f [ne1,ne2], ns)
  where (ne2, ns) = evalc pst (e2,s)
        Sub vars exps = ns
        ne1 = substitute vars exps e1

evalc_builtinEQaux :: ProState -> String -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinEQaux pst f [Comb _ f1 e1, Comb _ f2 e2] s
  | f1==f2                                      = evalc pst (makeSeqCon f e1 e2, s)
  | (f1=="False") && (f2=="failed") && f=="=="  = (Comb ConsCall "True" [], s)
  | (f1=="False") && (f2=="failed") && f=="=:=" = (Comb FuncCall "success" [],s)
  | (f1=="failed") && (f2=="False") && f=="=="  = (Comb ConsCall "True" [],s)
  | (f1=="failed") && (f2=="False") && f=="=:=" = (Comb FuncCall "success" [],s)
  | f=="=="                                     = (Comb ConsCall "False" [],s)
  | otherwise                                   = (Comb FuncCall "failed" [],s)
--the strange cases above are needed due to the current definition of Case Select
--(in particular, the case when the argument doesn't match any branch..)

evalc_builtinEQlit :: String -> Expr -> Expr -> Subst -> (Expr,Subst)
evalc_builtinEQlit f (Lit (Intc x)) (Lit (Intc y)) s
  | x==y && f=="=="   = (Comb ConsCall "True" [],s)
  | x==y && f=="=:="  = (Comb FuncCall "success" [],s)
  | x/=y && f=="=="   = (Comb ConsCall "False" [],s)
  | otherwise         = (Comb FuncCall "failed" [],s)


evalc_builtinEQ2 :: ProState -> String -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinEQ2 pst f [e1,e2] s
  | (isConsCall e1) && (isConsCall e2)  = evalc_builtinEQaux pst f [e1,e2] s
  | (isLitInt e1) && (isLitInt e2)      = evalc_builtinEQlit f e1 e2 s
  | isVar e1                            = evalc_builtinEQvar1 pst f [e1,e2] s
  | isVar e2                            = evalc_builtinEQvar2 pst f [e2,e1] s
  | (isLitInt e1) || (isConsCall e1)    = evalc_EQAux2 pst f e1 e2 s
  | otherwise                           = evalc_EQAux pst f e1 e2 s


evalc_builtinEQvar1 :: ProState -> String -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinEQvar1 pst f [Var x, e] s =
  if dataExp e && f=="=:="
     then evalc_builtinEQvarAux pst f x e s
     else evalcEQvar1_aux pst f (Var x) e s

evalcEQvar1_aux pst f v e s = 
  if ne==e 
     then (Comb FuncCall f [v,e], s)
     else (Comb FuncCall f [nv,ne], ns)
  where (ne, ns) = evalc pst (e,s)
        Sub vars exps = ns
        nv = substitute vars exps v

evalc_builtinEQvar2 :: ProState -> String -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinEQvar2 pst f [Var x, e] s =
  if dataExp e && f=="=:="
     then evalc_builtinEQvarAux pst f x e s
     else evalcEQvar2_aux pst f (Var x) e s

evalcEQvar2_aux pst f v e s = 
  if ne==e
     then (Comb FuncCall f [e,v], s)
     else (Comb FuncCall f [ne,nv], ns)
  where (ne, ns) = evalc pst (e,s)
        Sub vars exps = ns
        nv = substitute vars exps v

evalc_builtinEQvarAux :: ProState -> String -> Int -> Expr -> Subst -> (Expr,Subst)
evalc_builtinEQvarAux pst f x e s =
  evalc pst (Case Flex (Var x) (subs2branches e f), s)


subs2branches :: Expr -> String -> [BranchExpr]
subs2branches (Lit c) f = 
  if f=="==" 
     then [Branch (LPattern c) (Comb ConsCall "True" [])]
     else [Branch (LPattern c) (Comb FuncCall "success" [])]
subs2branches (Comb ConsCall c []) f =
  if f=="==" 
     then [Branch (Pattern c []) (Comb ConsCall "True" [])]
     else [Branch (Pattern c []) (Comb FuncCall "success" [])]
--PENDING: extend function above to cover arbitrary data terms...

--currently we only consider literals and constants.. TO DO..
dataExp :: Expr -> Bool
dataExp (Var _) = False
dataExp (Lit _) = True
dataExp (Comb ftype f args) = 
  args==[] && 
          (ftype==ConsCall || f=="success" || f=="failed")
-- if ftype==ConsCall then andL (map dataExp args) else False
dataExp (Apply _ _) = False
dataExp (Constr _ _) = False
dataExp (Or _ _) = False
dataExp (Case _ _ _) = False
dataExp (GuardedExpr _ _ _) = False
dataExp (Choice _) = False


isConsCall :: Expr -> Bool
isConsCall x = first (isConsCallAux x)
isConsCallAux :: Expr -> Bool
isConsCallAux (Comb ConsCall _ _) = True
isConsCallAux (Comb _ "success" _) = True
isConsCallAux (Comb _ "failed" _) = True
isConsCallAux _ = False

isConstant :: Expr -> Bool
isConstant x = first (isConstantAux x)
isConstantAux :: Expr -> Bool
isConstantAux (Comb ConsCall _ []) = True
isConstantAux (Comb _ "success" _) = True
isConstantAux (Comb _ "failed" _) = True
isConstantAux _ = False

makeSeqCon :: String -> [Expr] -> [Expr] -> Expr
makeSeqCon f [] [] = 
  if f=="==" then Comb ConsCall "True" []
             else Comb FuncCall "success" []
makeSeqCon f [e1] [e2] = Comb FuncCall f [e1,e2]
makeSeqCon f (e1:(e1b:es1)) (e2:(e2b:es2)) =
  if f=="==" 
     then (Comb FuncCall "&&" [(Comb FuncCall f [e1,e2]), (makeSeqCon f (e1b:es1) (e2b:es2))])
     else (Comb FuncCall "&" [(Comb FuncCall f [e1,e2]), (makeSeqCon f (e1b:es1) (e2b:es2))])

-------------------------------------------------------------------
-- IF THEN ELSE:

evalc_builtinITE :: ProState -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinITE pst [c,e1,e2] s
  | c==(Comb ConsCall "True" [])   = evalc pst (e1,s)
  | c==(Comb ConsCall "False" [])  = evalc pst (e2,s)
  | otherwise                      = evalc_ITEAux pst "if_then_else" [c,e1,e2] s

evalc_ITEAux pst _ [c,e1,e2] s = 
  if nc==c
     then (Comb FuncCall "if_then_else" [c,e1,e2], s)
     else (Comb FuncCall "if_then_else" [nc,ne1,ne2], ns)
  where (nc, ns) = evalc pst (c,s)
        Sub vars exps = ns
        ne1 = substitute vars exps e1
        ne2 = substitute vars exps e2

--evalc_builtinITE_aux pst [Comb FuncCall "==" [c1,c2], e1, e2] cc k
--  | (isLitInt c1) && (isLitInt c2) = if c1==c2 then evalc pst (e1,cc,k)
--                                               else evalc pst (e2,cc,k)
--  | otherwise = evalc_ITEAux pst "if_then_else" [Comb FuncCall "==" [c1,c2],e1,e2] cc k


--  | isVar c1 && isConstr c2 = (Comb FuncCall "if_then_else" 
--                              [Comb FuncCall "==" [c1,c2],
--                               substitute (vars2index [c1]) [c2] e1, 
--                               e2], cc, k)
--  | isVar c2 && isConstr c1 = (Comb FuncCall "if_then_else" 
--                              [Comb FuncCall "==" [c1,c2],
--                               substitute (vars2index [c2]) [c1] e1, 
--                               e2],cc,k)
--  | otherwise               = (Comb FuncCall "if_then_else" 
--                              [Comb FuncCall "==" [c1,c2], e1, e2], cc, k)

-------------------------------------------------------------------
-- SEQUENTIAL CONJUNCTION:

evalc_builtinAND :: ProState -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinAND pst [e1,e2] s
  | e1==(Comb ConsCall "True" [])    =  evalc pst (e2,s)
  | e1==(Comb FuncCall "success" []) =  evalc pst (e2,s)
  | e2==(Comb ConsCall "True" [])    =  evalc pst (e1,s)
  | e2==(Comb FuncCall "success" []) =  evalc pst (e1,s)
  | e1==(Comb ConsCall "False" [])   =  (Comb ConsCall "False" [], s)
  | e1==(Comb FuncCall "failed" [])  =  (Comb FuncCall "failed" [],s)
  | e2==(Comb ConsCall "False" [])   =  (Comb ConsCall "False" [],s)
  | e2==(Comb FuncCall "failed" [])  =  (Comb FuncCall "failed" [],s)
  | otherwise                        =  evalc_ANDAux pst e1 e2 s

evalc_ANDAux pst e1 e2 s = 
  if ne1==e1
     then (Comb FuncCall "&&" [e1,e2], s)
     else (Comb FuncCall "&&" [ne1,ne2], ns)
  where (ne1,ns) = evalc pst (e1,s)
        Sub vars exps = ns
        ne2 = substitute vars exps e2

 
--------------------------------------------------------------------
-- CONCURRENT CONJUNCTION:

evalc_builtinCAND :: ProState -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinCAND pst [e1,e2] s
  | e1==(Comb FuncCall "success" [])  = evalc pst (e2,s)
  | e2==(Comb FuncCall "success" [])  = evalc pst (e1,s)
  | e1==(Comb FuncCall "failed" [])   = (Comb FuncCall "failed" [],s)
  | e2==(Comb FuncCall "failed" [])   = (Comb FuncCall "failed" [],s)
  | otherwise                         = evalc_CANDAux pst e1 e2 s


evalc_CANDAux pst e1 e2 s = 
  if isSuspended e1 
    then evalc_CANDAux1 pst e1 e2 s
    else evalc_CANDAux2 pst e1 e2 s

evalc_CANDAux1 pst e1 e2 s = 
  if ne2==e2
     then (Comb FuncCall "&" [e1,e2], s)
     else (Comb FuncCall "&" [ne1,ne2], ns)
  where (ne2,ns) = evalc pst (e2,s)
        Sub vars exps = ns
        ne1 = substitute vars exps e1
        
evalc_CANDAux2 pst e1 e2 s = 
  if ne1==e1
     then (Comb FuncCall "&" [e1 ,e2], s)
     else (Comb FuncCall "&" [ne1 ,ne2], ns)
  where (ne1,ns) = evalc pst (e1 ,s)
        Sub vars exps = ns
        ne2 = substitute vars exps e2
        
isSuspended :: Expr -> Bool
isSuspended e = first (isSuspendedAux e)
isSuspendedAux :: Expr -> Bool
isSuspendedAux (Case Rigid (Var _) _) = True
isSuspendedAux (Comb FuncCall "&" es) = andL (map isSuspended es)
isSuspendedAux (Comb FuncCall "==" [e1,e2]) = (isVar e1) || (isVar e2)
isSuspendedAux (Comb FuncCall "=:=" [e1,e2]) = (isSuspended e1) || (isSuspended e2)
isSuspendedAux _ = False

-- mejora esta para que acepte suspensions! (solo va el colormap "malo"..)
--evalc_CANDAux pst e1 e2 cc k s = evalc pst (Comb FuncCall "&" [ne1,ne2], cc, nk, ns)
--  --where (ne1,_,nk, ns) = evalc pst (e1,cc,k,s)
--  where (ne1,_,nk,ns) = evalc pst (e1,cc,k,s)
--        Sub vars exps = ns
--        ne2 = substitute vars exps e2


-----------------------------------------------------------------------
-- ARITHMETIC AND RELATIONAL OPERATORS:

evalc_builtinARITH :: ProState -> String -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinARITH pst f [e1,e2] s =
  if (isLitInt e1) && (isLitInt e2)
     then evalc_builtinARITHaux f [e1,e2] s
     else if (isLitInt e2) then evalc_ARITHAux pst f e1 e2 s
             else evalc_ARITHAux2 pst f e1 e2 s

evalc_ARITHAux pst f e1 e2 s = 
  if ne1==e1
     then (Comb FuncCall f [e1,e2], s)
     else (Comb FuncCall f [ne1,e2], ns)
  where (ne1,ns) = evalc pst (e1,s)
evalc_ARITHAux2 pst f e1 e2 s = 
  if ne2==e2
     then (Comb FuncCall f [e1,e2], s)
     else (Comb FuncCall f [e1,ne2], ns)
  where (ne2,ns) = evalc pst (e2,s)


--PENDING: extend to floats and to more operators..


evalc_builtinARITHaux :: String -> [Expr] -> Subst -> (Expr,Subst)
evalc_builtinARITHaux f [(Lit (Intc e1)),(Lit (Intc e2))] s
  | (f=="*")   = (Lit (Intc (e1*e2)),s)
  | (f=="+")   = (Lit (Intc (e1+e2)),s)
  | (f=="-")   = (Lit (Intc (e1-e2)),s)
  | (f=="<")   = if e1<e2 then (Comb ConsCall "True" [],s) else (Comb ConsCall "False" [],s)
  | (f==">")   = if e1>e2 then (Comb ConsCall "True" [],s) else (Comb ConsCall "False" [],s)
  | (f=="<=")  = if e1<=e2 then (Comb ConsCall "True" [],s) else (Comb ConsCall "False" [],s)
  | (f==">=")  = if e1>=e2 then (Comb ConsCall "True" [],s) else (Comb ConsCall "False" [],s)

isLitInt :: Expr -> Bool
isLitInt e = first (isLitIntAux e)
isLitIntAux :: Expr -> Bool
isLitIntAux (Lit (Intc _)) = True
isLitIntAux _ = False


--check whether an expression is a variable:
isVar :: Expr -> Bool
isVar x = first (isVarAux x)
isVarAux:: Expr -> Bool
isVarAux (Var _) = True
isVarAux _ = False

--check whether an expression is a strict equality
isStrictEQ :: Expr -> Bool
isStrictEQ x = first (isStrictEQAux x)
isStrictEQAux :: Expr -> Bool
isStrictEQAux (Comb FuncCall "==" _) = True
isStrictEQAux _ = False

--check whether an expression is a literal:
isLit :: Expr -> Bool
isLit x = first (isLitAux x)
isLitAux:: Expr -> Bool
isLitAux (Lit _) = True
isLitAux _ = False

isConstr :: Expr -> Bool
isConstr e = first (isConstrAux e)
isConstrAux :: Expr -> Bool
isConstrAux (Var _) = True
isConstrAux (Lit _) = True
isConstrAux (Comb _ "success" _) = True
isConstrAux (Comb _ "failed" _) = True
isConstrAux (Comb ConsCall _ es) = andL (map isConstr es)
isConstrAux _ = False

isGroundConstr :: Expr -> Bool
isGroundConstr e = first (isGroundConstrAux e)
isGroundConstrAux :: Expr -> Bool
isGroundConstrAux (Var _) = False
isGroundConstrAux (Lit _) = True
isGroundConstrAux (Comb ConsCall _ es) = andL (map isGroundConstr es)
isGroundConstrAux _ = False


--conjunction of a list of boolean terms

andL:: [Bool] -> Bool
andL es = foldr (&&) True es

-- Get first solution
first g = head (findall (\x -> g =:= x))

-- (substitute vars exps expr) = expr[vars/exps]
-- i.e., replace all occurrences of vars by corresponding exps in the
-- expression expr
substitute vars exps expr = substituteAll vars exps 0 expr

-- (substituteAll vars exps base expr):
-- substitute all occurrences of variables by corresonding expressions:
-- * substitute all occurrences of var_i by exp_i in expr
--   (if vars=[var_1,...,var_n] and exps=[exp_1,...,exp_n])
-- * substitute all other variables (Var j) by (Var (base+j))
--
-- here we assume that the new variables in guards and case patterns
-- do not occur in the list "vars" of replaced variables!

substituteAll :: [Int] -> [Expr] -> Int -> Expr -> Expr
substituteAll vars exps b (Var i) = replaceVar vars exps i
  where replaceVar [] [] var = Var (b+var)
        replaceVar (v:vs) (e:es) var = if v==var then e
                                                 else replaceVar vs es var
substituteAll _  _  _ (Lit l) = Lit l
substituteAll vs es b (Comb combtype c exps) =
                 Comb combtype c (map (substituteAll vs es b) exps)
substituteAll vs es b (Apply e1 e2) =
                 Apply (substituteAll vs es b e1) (substituteAll vs es b e2)
substituteAll vs es b (Constr vars e) =
                 Constr (map (+b) vars) (substituteAll vs es b e)
substituteAll vs es b (Or e1 e2) =
                 Or (substituteAll vs es b e1) (substituteAll vs es b e2)
substituteAll vs es b (Case ctype e cases) =
   Case ctype (substituteAll vs es b e) (map (substituteAllCase vs es b) cases)
substituteAll vs es b (GuardedExpr vars c e) =
   GuardedExpr (map (+b) vars) (substituteAll vs es b c)
                               (substituteAll vs es b e)
substituteAll vs es b (Choice e) = Choice (substituteAll vs es b e)

substituteAllCase vs es b (Branch (Pattern l pvs) e) =
                 Branch (Pattern l (map (+b) pvs)) (substituteAll vs es b e)
substituteAllCase vs es b (Branch (LPattern l) e) =
                 Branch (LPattern l) (substituteAll vs es b e)

blanks b = take b (repeat ' ')


--end of program

